@@ -32,6 +32,8 @@
 #define PN_PROTO_PHONET		1
 #define PN_PROTO_PIPE		2
 
+#define SOL_PNPIPE		275
+
 #define PNPIPE_ENCAP		1
 #define PNPIPE_IFINDEX		2
 
@@ -39,6 +41,8 @@
 #define PNPIPE_ENCAP_IP		1
 
 #define SIOCPNGETOBJECT		(SIOCPROTOPRIVATE + 0)
+#define SIOCPNADDRESOURCE	(SIOCPROTOPRIVATE + 14)
+#define SIOCPNDELRESOURCE	(SIOCPROTOPRIVATE + 15)
 
 struct sockaddr_pn {
 	sa_family_t spn_family;
