@@ -203,7 +203,6 @@ static void cfun_set_on_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		int i;
 
 		for (i = 0; i < NUM_DLC; i++) {
-			g_at_chat_shutdown(data->dlcs[i]);
 			g_at_chat_unref(data->dlcs[i]);
 			data->dlcs[i] = NULL;
 		}
@@ -410,7 +409,6 @@ static int calypso_disable(struct ofono_modem *modem)
 	DBG("");
 
 	for (i = 0; i < NUM_DLC; i++) {
-		g_at_chat_shutdown(data->dlcs[i]);
 		g_at_chat_unref(data->dlcs[i]);
 		data->dlcs[i] = NULL;
 	}
@@ -431,12 +429,16 @@ static int calypso_disable(struct ofono_modem *modem)
 static void calypso_pre_sim(struct ofono_modem *modem)
 {
 	struct calypso_data *data = ofono_modem_get_data(modem);
+	struct ofono_sim *sim;
 
 	DBG("");
 
 	ofono_devinfo_create(modem, 0, "atmodem", data->dlcs[AUX_DLC]);
-	ofono_sim_create(modem, 0, "atmodem", data->dlcs[AUX_DLC]);
+	sim = ofono_sim_create(modem, 0, "atmodem", data->dlcs[AUX_DLC]);
 	ofono_voicecall_create(modem, 0, "calypsomodem", data->dlcs[VOICE_DLC]);
+
+	if (sim)
+		ofono_sim_inserted_notify(sim, TRUE);
 }
 
 static void calypso_post_sim(struct ofono_modem *modem)
